#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_satellite_populate
----------------------------------

Tests for `satellite_populate` module.
"""

from click.testing import CliRunner
from satellite_populate import commands

runner = CliRunner()


def test_invoked_with_no_datafile():
    result = runner.invoke(commands.main)
    assert result.exit_code == 2
    assert 'Missing argument "datafile"' in result.output


def test_help_output():
    help_result = runner.invoke(commands.main, ['--help'])
    assert help_result.exit_code == 0
    assert 'https://satellite-populate.readthedocs.io' in help_result.output


def test_raises_with_invalid_path():
    result = runner.invoke(commands.main, ['/foo/baz/zaz.yaml'])
    assert result.exit_code == -1


def test_raises_with_invalid_extension():
    result = runner.invoke(commands.main, ['/foo/baz/zaz.txt'])
    assert result.exit_code == -1


def test_validate_raise_validation_exit_status():
    yaml = (
        "actions: ["
        "{'action': 'assertion', 'operator': 'eq', 'data': [1, 2]}"
        "]"
    )
    mode = '--mode=validate'
    result = runner.invoke(commands.main, [yaml, mode])
    assert result.exit_code == 1
    assert 'System entities did not validated!' in result.output
    assert '1 is NOT eq to 2' in result.output


def test_populate_do_not_raise_validation_exit_status():
    yaml = (
        "actions: ["
        "{'action': 'assertion', 'operator': 'eq', 'data': [1, 2]}"
        "]"
    )
    result = runner.invoke(commands.main, [yaml])
    assert result.exit_code == 0
    assert 'System entities did not validated!' not in result.output
    assert '1 is NOT eq to 2' in result.output



# coding: utf-8
"""
This module contains commands to interact with satellite populator
and validator.

Commands included:

satellite-populate
--------

A command to populate the system based in an YAML file describing the
entities::

    $ satellite-populate file.yaml -h myhost.com -o /tmp/validation.yaml

validate
--------

A command to validate the system based in an validation file generated by
the populate or a YAML file with mode: validation::

    $ satellite-populate /tmp/validation.yaml

Use :code:`$ satellite-populate --help` for more info
"""
import sys

import click

from satellite_populate.main import populate


def execute_populate(datafile, verbose, output, mode, scheme, port, hostname,
                     username, password):
    """Populate using the data described in `datafile`:"""
    result = populate(
        datafile,
        verbose=verbose,
        output=output,
        mode=mode,
        scheme=scheme,
        port=port,
        hostname=hostname,
        username=username,
        password=password
    )

    result.logger.info(
        "{0} entities already existing in the system".format(
            len(result.found)
        )
    )

    if result.mode == 'populate':
        result.logger.info(
            "{0} entities were created in the system".format(
                len(result.created)
            )
        )

    if result.assertion_errors:
        for error in result.assertion_errors:
            data = error['data']
            result.logger.error(
                'assertion: %s is NOT %s to %s',
                data['value'], error['operator'], data['other']
            )
        if result.mode == 'validate':
            sys.exit("System entities did not validated!")

    if result.validation_errors:
        for error in result.validation_errors:
            result.logger.error(error['message'])
            result.logger.error(error['search'])
        if result.mode == 'validate':
            sys.exit("System entities did not validated!")
    else:
        if result.mode == 'validate':
            result.logger.info("System Validated!!!")


VERBOSE_HELP = """
envvar:POPULATE_VERBOSE
-v(1):DEBUG
-vv(2):INFO
-vvv(3):WARNING
-vvvv(4):ERROR
-vvvvv(5):CRITICAL
"""


@click.command()
@click.argument(
    'datafile',
    required=True,
    type=click.Path(),
    envvar='POPULATE_DATAFILE'
)
@click.option(
    '-v', '--verbose',
    count=True,
    envvar='POPULATE_VERBOSE',
    help=VERBOSE_HELP
)
@click.option(
    '-o', '--output',
    type=click.Path(),
    default=None,
    envvar='POPULATE_OUTPUT',
    help="""envvar:POPULATE_OUTPUT
    Filepath to output the validation file e.g: -o=validation_123.yaml
    """
)
@click.option(
    '-m', '--mode',
    type=click.Choice(choices=('populate', 'validate')),
    default=None,
    envvar='POPULATE_MODE',
    help="""envvar:POPULATE_MODE
    valid modes are populate and validate"""
)
@click.option('--scheme', default=None, envvar='POPULATE_SCHEMA',
              help="""envvar:POPULATE_SCHEMA\noptional url schema""")
@click.option('--port', default=None, envvar='POPULATE_PORT', type=int,
              help="""envvar:POPULATE_PORT\noptional url port""")
@click.option('-h', '--hostname', default=None, envvar='POPULATE_HOSTNAME',
              help="""envvar:POPULATE_HOSTNAME\nSatellite URL:server.com""")
@click.option('-u', '--username', default=None, envvar='POPULATE_USERNAME',
              help="""envvar:POPULATE_USERNAME\nAdmin user""")
@click.option('-p', '--password', default=None, envvar='POPULATE_PASSWORD',
              help="""envvar:POPULATE_PASSWORD\nAdmin Password""")
def main(datafile, verbose, output, mode, scheme, port, hostname, username,
         password):
    """Populates or validates Satellite entities. Full documentation can be
    found on: https://satellite-populate.readthedocs.io\n
        $ satellite-populate test_data.yaml -vv\n
        $ satellite-populate test_data.yaml -vv --mode=validate\n
        $ satellite-populate test_data.yaml -vv -h myserver.com
    """
    execute_populate(datafile, verbose, output, mode, scheme, port,
                     hostname, username, password)


if __name__ == "__main__":
    main()



# coding: utf8
"""Point of entry for populate and validate used in scripts"""
import json
from collections import OrderedDict
from pprint import pformat

import import_string
import os
import yaml

from satellite_populate.constants import DEFAULT_CONFIG
from satellite_populate.utils import remove_keys, remove_nones


def setup_yaml():
    """Set YAML to use OrderedDict
    http://stackoverflow.com/a/8661021"""
    represent_dict_order = lambda self, data: self.represent_mapping(  # noqa
        'tag:yaml.org,2002:map', data.items()
    )
    yaml.add_representer(OrderedDict, represent_dict_order)


setup_yaml()


def load_data(datafile):
    """Loads YAML file as a dictionary"""
    if datafile.endswith(('.yml', '.yaml', 'json')):
        with open(datafile) as opened_file:
            data = yaml.load(opened_file)
            data['input_filename'] = datafile
            return data
    return yaml.load(datafile)


def get_populator(data, **kwargs):
    """Gets an instance of populator dynamically"""
    if not isinstance(data, dict):
        data = load_data(data)

    if not isinstance(data, dict):
        raise ValueError(
            "Data must be a valid filepath, dict, json or YAML string"
        )

    config = DEFAULT_CONFIG.copy()
    config.update(remove_nones(data.get('config', {})))
    config.update(remove_nones(kwargs))
    populator_name = config['populator']
    populator_module_name = config['populators'][populator_name]['module']
    populator_class = import_string(populator_module_name)
    populator = populator_class(data=data, config=config)
    return populator


def populate(data, **kwargs):
    """Loads and execute populator in populate mode"""
    kwargs = remove_nones(kwargs)
    populator = get_populator(data, **kwargs)
    populator.execute()
    populator.logger.info("%s finished!", populator.mode)
    if kwargs.get('output'):
        save_rendered_data(populator, kwargs['output'])
    return populator


def default_context_wrapper(result):
    """Takes the result of populator and keeps only useful data
    e.g. in decorators context.registered_name, context.config.verbose and
    context.vars.admin_username will all be available.
    """
    context = result.registry
    context.config = result.config
    context.vars = result.vars
    return context


def save_rendered_data(result, filepath):
    """Save the result of rendering in a new file to be used for
    validation"""
    file_format = os.path.splitext(filepath)[-1]
    if file_format not in ('.json', '.yml', '.yaml', '.py'):
        raise ValueError("Invalid filename extension")

    data = OrderedDict({
        'input_filename': result.input_filename,
        'config': dict(result.config),
        'vars': dict(result.vars),
        'actions': result.rendered_actions
    })

    data['config']['mode'] = 'validate'
    data['config'] = remove_keys(data['config'], 'output')
    # should remove username, password, hostname ?

    data['config'] = remove_nones(data['config'])

    directory, filename = os.path.split(filepath)

    if not directory:
        directory = os.path.dirname(result.input_filename)

    if not filename.startswith('validation'):
        filename = "validation_{0}".format(filename)

    filepath = os.path.join(directory, filename)

    with open(filepath, 'w') as output:
        if file_format in ('.yml', '.yaml'):
            output.write(yaml.dump(data))
        elif file_format == '.json':
            output.write(json.dumps(data, indent=4))
        elif file_format == '.py':
            output.write("data = {0}".format(pformat(dict(data))))

    result.logger.debug("Validation data saved in %s", filepath)



"""decorators for populate feature"""
from functools import wraps

from satellite_populate.main import populate, default_context_wrapper


def populate_with(data, context_name=None,
                  context_wrapper=default_context_wrapper, **extra_options):
    """To be used in test cases as a decorator

    Having a data_file like::

        actions:
          - model: Organization
            register: organization_1
            data:
              name: My Org

    Then you can use in decorators::

        @populate_with('file.yaml')
        def test_case_(self):
            'My Org exists in system test anything here'

    And getting the populated entities inside the test_case::

        @populate_with('file.yaml', context_name='my_context')
        def test_case_(self, my_context=None):
            assert my_context.organization_1.name == 'My Org'

    You can also set a customized context wrapper to the
    context_wrapper argument::

        def my_custom_context_wrapper(result):
            # create an object using result
            my_context = MyResultContext(result)
            return my_context

        @populate_with('file.yaml', context_name='my_context',
                       content_wrapper=my_custom_context_wrapper)
        def test_case_(self, my_context=None):
            # assert with some expression using my_context object returned
            # my_custom_context_wrapper
            assert some_expression

    NOTE::

        That is important that ``context_name`` argument always be declared
        using either a default value ``my_context=None`` or handle in
        ``**kwargs`` Otherwise ``py.test`` may try to use this as a fixture
         placeholder.

        if context_wrapper is set to None, my_context will be the pure
        unmodified result of populate function.

    """

    def decorator(func):
        """Wrap test method"""

        @wraps(func)
        def wrapper(*args, **kwargs):
            """decorator wrapper"""
            if not 'verbose' in extra_options:
                # set default verbose to -v so we get all messages in test log
                extra_options['verbose'] = 1
            result = populate(data, **extra_options)
            if context_name is not None:
                if context_wrapper:
                    context = context_wrapper(result)
                else:
                    context = result

                kwargs[context_name] = context

            return func(*args, **kwargs)

        return wrapper

    return decorator



# coding: utf-8
"""
Tests for satellite_populate module and commands
"""
import pytest
from unittest2 import TestCase
from satellite_populate import populate_with


data_in_dict = {
    'actions': [
        {
            'action': 'register',
            'data': {
                'project': 'Satellite'
            }
        }
        # CRUD ACTIONS DEMANDS MOCK OF API
        # {
        #     'model': 'Organization',
        #     'register': 'organization_1',
        #     'data': {
        #         'name': 'My Organization 1',
        #         'label': 'my_organization_1'
        #     }
        # },
        # {
        #     'model': 'Organization',
        #     'register': 'organization_2',
        #     'data': {
        #         'name': 'My Organization 2',
        #         'label': 'my_organization_2'
        #     }
        # }
    ]
}

data_in_string = """
actions:
  - action: register
    data:
      name: Michael
      lastname: Scott
      company: Dunder Mifflin
      fake_repo:
        from_object: nailgun.entities._FAKE_YUM_REPO
      random:
        from_factory:
          alpha: 10
      should_not_exist: "This will be unregistered below"

  - action: assertion
    operation: eq
    data:
      - Hello
      - Hello
    register: hello_equals_hello

  - action: unregister
    data:
      - should_not_exist
"""


class CustomizedResultContext(object):
    """a simple customized result context wrapper object"""

    def __init__(self, result):
        self.__result = result

    @classmethod
    def wrapper(cls, result):
        return cls(result)

    @property
    def result(self):
        return self.__result

    @property
    def registry(self):
        return self.result.registry

    @property
    def vars(self):
        return self.result.vars

    @property
    def config(self):
        return self.result.config

    def __getattr__(self, name):
        return getattr(self.result.registry, name)


@populate_with(data_in_dict, context_name='context', verbose=1)
def test_data_in_dict(context=None):
    """a test with populated data"""
    assert context.project == "Satellite"


@populate_with(data_in_dict, context_name='context',
               context_wrapper=CustomizedResultContext.wrapper,
               verbose=0)
def test_data_in_dict_and_customized_wrapper(context=None):
    """a test with populated data and customized wrapper"""
    assert isinstance(context, CustomizedResultContext)
    assert context.project == context.registry.project
    assert context.project == "Satellite"


@populate_with(data_in_string, context_name='context', verbose=1)
def test_data_in_string(**kwargs):
    """a test with populated data"""
    context = kwargs['context']
    assert context.name == 'Michael'
    assert context.lastname == 'Scott'
    assert context.company == 'Dunder Mifflin'
    fake_repo = 'http://inecas.fedorapeople.org/fakerepos/zoo3/'
    assert context.fake_repo == fake_repo
    assert len(context.random) == 10
    assert context.hello_equals_hello is True
    with pytest.raises(KeyError):
        assert context.should_not_exist == "This will be unregistered below"


class MyTestCase(TestCase):
    """
    THis test populates data in setUp and also in individual tests
    """
    @populate_with(data_in_string, context_name='context')
    def setUp(self, context=None):
        self.context = context

    def test_with_setup_data(self):
        context = self.context
        self.assertEqual(context.name, 'Michael')
        self.assertEqual(context.lastname, 'Scott')
        self.assertEqual(context.company, 'Dunder Mifflin')
        fake_repo = 'http://inecas.fedorapeople.org/fakerepos/zoo3/'
        self.assertEqual(context.fake_repo, fake_repo)
        self.assertEqual(len(context.random), 10)
        self.assertTrue(context.hello_equals_hello)

        with self.assertRaises(KeyError):
            self.assertEqual(context.should_not_exist,
                             "This will be unregistered below")

    @populate_with(data_in_dict, context_name='test_context')
    def test_with_isolated_data(self, test_context=None):
        self.assertEqual(
            test_context.project, "Satellite"
        )
